from typing import Optional

from pybotx_smartapp_rpc.exception_handlers import default_exception_handler
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    ExceptionHandler,
    ExceptionHandlerDict,
    HandlerWithArgs,
    RPCArgsBaseModel,
    RPCResponse,
)


class ExceptionMiddleware:
    def __init__(
        self,
        exception_handlers: Optional[ExceptionHandlerDict] = None,
    ) -> None:
        self._exception_handlers: ExceptionHandlerDict = exception_handlers or {}

    async def __call__(
        self,
        smartapp: SmartApp,
        rpc_arguments: RPCArgsBaseModel,
        call_next: HandlerWithArgs,
    ) -> RPCResponse:
        try:
            rpc_result = await call_next(smartapp, rpc_arguments)
        except Exception as exc:
            exception_handler = self._get_exception_handler(exc)
            try:  # noqa: WPS505
                return await exception_handler(exc, smartapp)
            except Exception as error_handler_exc:
                return await default_exception_handler(error_handler_exc, smartapp)

        return rpc_result

    def _get_exception_handler(self, exc: Exception) -> ExceptionHandler:
        for exc_cls in type(exc).mro():
            handler = self._exception_handlers.get(exc_cls)
            if handler:
                return handler

        raise RuntimeError("No handler for Exception found.")  # pragma: no cover



from types import SimpleNamespace
from typing import Any, Dict, List, Optional
from uuid import UUID

from pybotx import Bot, File, SmartAppEvent
from pybotx.missing import Missing, Undefined


class SmartApp:
    def __init__(
        self,
        bot: Bot,
        bot_id: UUID,
        chat_id: UUID,
        event: Optional[SmartAppEvent] = None,
    ) -> None:
        self.bot = bot
        self.event = event

        self.bot_id = bot_id
        self.chat_id = chat_id

        self.state = SimpleNamespace()

    async def send_event(
        self,
        rpc_result: Any,
        files: Optional[List[File]] = None,
        encrypted: bool = True,
    ) -> None:
        await self.bot.send_smartapp_event(
            bot_id=self.bot_id,
            chat_id=self.chat_id,
            data={
                "status": "ok",
                "type": "smartapp_rpc",
                "result": rpc_result,
            },
            files=files or [],
            encrypted=encrypted,
        )

    async def send_push(self, counter: int, body: Missing[str] = Undefined) -> None:
        await self.bot.send_smartapp_notification(
            bot_id=self.bot_id,
            chat_id=self.chat_id,
            smartapp_counter=counter,
            body=body,
        )

    async def send_custom_push(
        self,
        title: str,
        body: str,
        meta: Missing[Dict[str, Any]] = Undefined,
        wait_callback: bool = True,
        callback_timeout: Optional[float] = None,
    ) -> UUID:
        return await self.bot.send_smartapp_custom_notification(
            bot_id=self.bot_id,
            group_chat_id=self.chat_id,
            title=title,
            body=body,
            meta=meta,
            wait_callback=wait_callback,
            callback_timeout=callback_timeout,
        )



from typing import Any, Dict, Union

from pydantic import BaseModel, Field


class RPCError(BaseModel):
    reason: str
    id: str
    meta: Union[Dict[str, Any], BaseModel] = Field(default_factory=dict)



from pybotx_smartapp_rpc.exceptions import RPCErrorExc
from pybotx_smartapp_rpc.models.errors import RPCError
from pybotx_smartapp_rpc.models.request import RPCArgsBaseModel
from pybotx_smartapp_rpc.models.responses import (
    RPCErrorResponse,
    RPCResponseBaseModel,
    RPCResultResponse,
)
from pybotx_smartapp_rpc.router import RPCRouter
from pybotx_smartapp_rpc.rpc import SmartAppRPC
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    Handler,
    HandlerWithArgs,
    HandlerWithoutArgs,
    RPCResponse,
)

__all__ = (
    "Handler",
    "HandlerWithArgs",
    "HandlerWithoutArgs",
    "RPCArgsBaseModel",
    "RPCError",
    "RPCErrorExc",
    "RPCErrorResponse",
    "RPCResponse",
    "RPCResponseBaseModel",
    "RPCResultResponse",
    "RPCRouter",
    "SmartApp",
    "SmartAppRPC",
)



from dataclasses import dataclass, field
from typing import Any, Dict, Generic, List, TypeVar, Union

from pybotx import File
from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError

from pybotx_smartapp_rpc.models.errors import RPCError

_JsonableResultType = Union[float, int, str, bool, List, Dict]
JsonableResultType = TypeVar("JsonableResultType", bound=_JsonableResultType)

_ResultType = Union[BaseModel, _JsonableResultType]
ResultType = TypeVar("ResultType", bound=_ResultType)


class RPCResponseBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True


@dataclass
class RPCResultResponse(Generic[ResultType]):
    result: ResultType
    files: List[File] = field(default_factory=list)
    encrypted: bool = True

    def jsonable_dict(self) -> Dict[str, Any]:
        return {
            "status": "ok",
            "type": "smartapp_rpc",
            "result": self.jsonable_result(),
        }

    def jsonable_result(self) -> _JsonableResultType:
        if isinstance(self.result, BaseModel):
            return self.result.dict(by_alias=True)

        return self.result


@dataclass
class RPCErrorResponse:
    errors: List[RPCError]
    files: List[File] = field(default_factory=list)
    encrypted: bool = True

    def jsonable_dict(self) -> Dict[str, Any]:
        return {
            "status": "error",
            "type": "smartapp_rpc",
            "errors": self.jsonable_errors(),
        }

    def jsonable_errors(self) -> List[Dict[str, Any]]:
        return [error.dict() for error in self.errors]


def build_invalid_rpc_request_error_response(
    exc: ValidationError,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason=f"Invalid RPC request: {error['msg']}",
                id=error["type"].split(".")[0].upper(),
                meta={"field": error["loc"][0]},
            )
            for error in exc.errors()
        ],
    )


def build_invalid_rpc_args_error_response(
    exc: ValidationError,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason=error["msg"],
                id=error["type"].split(".")[0].upper(),
                meta={"location": error["loc"]},
            )
            for error in exc.errors()
        ],
    )


def build_method_not_found_error_response(
    method: str,
) -> RPCErrorResponse:
    return RPCErrorResponse(
        errors=[
            RPCError(
                reason="Method not found",
                id="METHOD_NOT_FOUND",
                meta={"method": method},
            ),
        ],
    )



from typing import Any, Dict, Literal

from pydantic import BaseModel


class RPCArgsBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True


class RPCRequest(BaseModel):
    method: str
    type: Literal["smartapp_rpc"]
    params: Dict[str, Any] = {}



from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import Dict, List, Optional, Union

from pydantic.fields import ModelField

from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import (
    Handler,
    HandlerWithArgs,
    Middleware,
    RPCArgsBaseModel,
    RPCResponse,
)


@dataclass
class RPCMethod:
    handler: Handler
    middlewares: List[Middleware]
    response_field: ModelField
    arguments_field: Optional[ModelField] = None
    tags: List[Union[str, Enum]] = field(default_factory=list)
    errors: Dict[str, dict] = field(default_factory=dict)
    errors_models: Dict[str, ModelField] = field(default_factory=dict)
    include_in_schema: bool = True

    async def __call__(
        self,
        smartapp: SmartApp,
        rpc_args: RPCArgsBaseModel,
    ) -> RPCResponse:
        # loop in reverse order
        # if middlewares = [m1, m2] and method.middlewares = [m3, m4]
        # then stack will be m1(m2(m3(m4(handler()))))
        handler: HandlerWithArgs = self.handler  # type: ignore
        for middleware in self.middlewares[::-1]:
            part = partial(middleware, call_next=handler)  # type: ignore
            handler = part

        return await handler(smartapp, rpc_args)



from enum import Enum
from typing import Any, Dict, List, Optional, Set, Type, Union

from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.schema import (
    field_schema,
    get_flat_models_from_fields,
    model_process_schema,
)

from pybotx_smartapp_rpc import RPCRouter
from pybotx_smartapp_rpc.models.method import RPCMethod

REF_PREFIX = "#/components/schemas/"


def deep_dict_update(
    destination_dict: Dict[Any, Any],
    source_dict: Dict[Any, Any],
) -> None:
    for key in source_dict.keys():
        if (
            key in destination_dict
            and isinstance(destination_dict[key], dict)
            and isinstance(source_dict[key], dict)
        ):
            deep_dict_update(destination_dict[key], source_dict[key])
        else:
            destination_dict[key] = source_dict[key]


def get_rpc_flat_models_from_routes(
    router: RPCRouter,
) -> Set[Union[Type[BaseModel], Type[Enum]]]:
    body_fields_from_routes: List[ModelField] = []
    responses_from_routes: List[ModelField] = []

    for method_name in router.rpc_methods.keys():
        if not router.rpc_methods[method_name].include_in_schema:
            continue

        if router.rpc_methods[method_name].arguments_field:
            body_fields_from_routes.append(
                router.rpc_methods[method_name].arguments_field,  # type: ignore
            )
        if router.rpc_methods[method_name].response_field:
            responses_from_routes.append(router.rpc_methods[method_name].response_field)

        if router.rpc_methods[method_name].errors_models:
            responses_from_routes.extend(
                router.rpc_methods[method_name].errors_models.values(),
            )

    return get_flat_models_from_fields(
        body_fields_from_routes + responses_from_routes,
        known_models=set(),
    )


def get_rpc_model_definitions(
    *,
    flat_models: Set[Union[Type[BaseModel], Type[Enum]]],
    model_name_map: Dict[Union[Type[BaseModel], Type[Enum]], str],
) -> Dict[str, Any]:
    definitions: Dict[str, Dict[str, Any]] = {}
    for model in flat_models:
        m_schema, m_definitions, m_nested_models = model_process_schema(
            model,
            model_name_map=model_name_map,
            ref_prefix=REF_PREFIX,
        )
        definitions.update(m_definitions)
        model_name = model_name_map[model]
        if "description" in m_schema:
            m_schema["description"] = m_schema["description"].split("\f")[0]
        definitions[model_name] = m_schema
    return definitions


def get_openapi_operation_rpc_args(
    *,
    body_field: Optional[ModelField],
    model_name_map: Dict[Union[Type[BaseModel], Type[Enum]], str],
) -> Optional[Dict[str, Any]]:
    if not body_field:
        return None

    body_schema, _, _ = field_schema(
        body_field,
        model_name_map=model_name_map,
        ref_prefix=REF_PREFIX,
    )
    request_media_type = "application/json"
    required = body_field.required
    request_body_oai: Dict[str, Any] = {}
    if required:
        request_body_oai["required"] = required

    request_media_content: Dict[str, Any] = {"schema": body_schema}
    request_body_oai["content"] = {request_media_type: request_media_content}
    return request_body_oai


def get_openapi_rpc_metadata(*, name: str, route: RPCMethod) -> Dict[str, Any]:
    operation: Dict[str, Any] = {
        "summary": route.handler.__name__.replace(".", " ").replace("_", " ").title(),
        "description": route.handler.__doc__,
        "operationId": (
            f"rpc_{name.replace('.', '_').replace(':', '_').replace('-', '_').lower()}"
        ),
    }

    if route.tags:
        operation["tags"] = route.tags

    return operation


def get_rpc_openapi_path(  # noqa: WPS231
    *,
    method_name: str,
    route: RPCMethod,
    model_name_map: Dict[Union[Type[BaseModel], Type[Enum]], str],
    security_scheme: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Taken from FastAPI."""
    path = {}

    operation = get_openapi_rpc_metadata(name=method_name, route=route)

    request_body_oai = get_openapi_operation_rpc_args(
        body_field=route.arguments_field,
        model_name_map=model_name_map,
    )
    if request_body_oai:
        operation["requestBody"] = request_body_oai

    # - Successful response -
    response_schema, _, _ = field_schema(
        route.response_field,
        model_name_map=model_name_map,
        ref_prefix=REF_PREFIX,
    )

    operation.setdefault("responses", {}).setdefault("ok", {}).update(
        {
            "description": "Successful response. **result** field:",
            "content": {"application/json": {"schema": response_schema}},
        }
    )

    # - Errors -
    if route.errors:
        operation_errors = operation.setdefault("responses", {})
        for error_status_code, error_response in route.errors.items():
            process_response: Dict[str, Any] = {}
            openapi_response = operation_errors.setdefault(str(error_status_code), {})

            if route.errors_models and (
                field := route.errors_models[error_status_code]  # noqa: WPS332
            ):
                error_field_schema, _, _ = field_schema(
                    field,
                    model_name_map=model_name_map,
                    ref_prefix=REF_PREFIX,
                )
                error_schema = (
                    process_response.setdefault("content", {})
                    .setdefault("application/json", {})
                    .setdefault("schema", {})
                )
                deep_dict_update(error_schema, error_field_schema)

            description = error_response["description"] or "Error"
            deep_dict_update(openapi_response, process_response)
            openapi_response["description"] = f"**Error**: {description}"

    if security_scheme:
        operation.setdefault("security", []).append(security_scheme)

    path["post"] = operation
    return path



import inspect
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseConfig
from pydantic.error_wrappers import ValidationError
from pydantic.fields import ModelField

from pybotx_smartapp_rpc import RPCError
from pybotx_smartapp_rpc.empty_args import EmptyArgs
from pybotx_smartapp_rpc.middlewares.empty_args_middleware import empty_args_middleware
from pybotx_smartapp_rpc.models.method import RPCMethod
from pybotx_smartapp_rpc.models.request import RPCRequest
from pybotx_smartapp_rpc.models.responses import (
    ResultType,
    build_invalid_rpc_args_error_response,
    build_method_not_found_error_response,
)
from pybotx_smartapp_rpc.smartapp import SmartApp
from pybotx_smartapp_rpc.typing import Handler, Middleware, RPCResponse


class RPCRouter:
    def __init__(  # noqa: WPS234
        self,
        middlewares: Optional[List[Middleware]] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        include_in_schema: bool = True,
        errors: Optional[List[Type[RPCError]]] = None,
    ) -> None:
        self.rpc_methods: Dict[str, RPCMethod] = {}
        self.middlewares: List[Middleware] = middlewares or []
        self.tags: List[Union[str, Enum]] = tags or []
        self.include_in_schema = include_in_schema
        self.errors: List[Type[RPCError]] = errors or []

    def method(  # noqa: WPS211, WPS234
        self,
        rpc_method_name: str,
        middlewares: Optional[List[Middleware]] = None,
        return_type: Optional[Type[ResultType]] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        errors: Optional[List[Type[RPCError]]] = None,
        include_in_schema: bool = True,
    ) -> Callable[[Handler], Handler]:
        if rpc_method_name in self.rpc_methods:
            raise ValueError(f"RPC method {rpc_method_name} already registered!")

        method_and_router_middlewares = self.middlewares + (middlewares or [])
        method_and_router_middlewares += [empty_args_middleware]

        current_tags = self.tags.copy()
        if tags:
            current_tags.extend(tags)

        def decorator(handler: Handler) -> Handler:
            arg_field, response_field = self._get_args_and_return_field(
                handler,
                return_type,
            )
            errors_fields, errors_models = self._get_error_fields_and_models(
                self.errors + (errors or []),
            )

            self.rpc_methods[rpc_method_name] = RPCMethod(
                handler=handler,
                middlewares=method_and_router_middlewares,
                response_field=response_field,
                arguments_field=arg_field,
                tags=current_tags,
                errors=errors_fields,
                errors_models=errors_models,
                include_in_schema=include_in_schema and self.include_in_schema,
            )

            return handler

        return decorator

    async def perform_rpc_request(
        self,
        smartapp: SmartApp,
        rpc_request: RPCRequest,
    ) -> RPCResponse:
        rpc_method = self.rpc_methods.get(rpc_request.method)
        if not rpc_method:
            return build_method_not_found_error_response(rpc_request.method)

        if rpc_method.arguments_field:
            try:
                args = rpc_method.arguments_field.type_(**rpc_request.params)
            except ValidationError as invalid_rpc_args_exc:
                return build_invalid_rpc_args_error_response(invalid_rpc_args_exc)
        else:
            args = EmptyArgs()

        return await rpc_method(smartapp, args)

    def include(self, *routers: "RPCRouter") -> None:
        for router in routers:
            self.include_router(router)

    def include_router(self, router: "RPCRouter") -> None:
        already_exist_handlers = self.rpc_methods.keys() & router.rpc_methods.keys()
        if already_exist_handlers:
            raise ValueError(
                f"RPC methods {already_exist_handlers} already registered!",
            )

        errors_fields, errors_models = self._get_error_fields_and_models(self.errors)

        for rpc_method_name, rpc_method in router.rpc_methods.items():
            rpc_method.middlewares = self.middlewares + rpc_method.middlewares
            rpc_method.errors = {**errors_fields, **rpc_method.errors}
            rpc_method.errors_models = {**errors_models, **rpc_method.errors_models}
            self.rpc_methods[rpc_method_name] = rpc_method

    def _get_args_and_return_field(
        self,
        handler: Handler,
        return_type: Optional[Type[ResultType]] = None,
    ) -> Tuple[Optional[ModelField], ModelField]:
        signature = inspect.signature(handler)
        return_annotation = signature.return_annotation
        if hasattr(return_annotation, "__args__"):  # noqa: WPS421
            response_type = return_annotation.__args__[0]
        else:
            response_type = None

        if return_type:
            response_type = return_type

        response_field = ModelField(
            name=f"Response_{handler.__name__}",
            type_=response_type,
            model_config=BaseConfig,
            class_validators={},
        )

        args_annotations = [arg[1].annotation for arg in signature.parameters.items()]
        if len(args_annotations) >= 2:
            arg_field = ModelField(
                name=str(args_annotations[1].__name__),
                type_=args_annotations[1],
                model_config=BaseConfig,
                class_validators={},
            )
        else:
            arg_field = None

        return arg_field, response_field

    def _get_error_fields_and_models(  # noqa: WPS234
        self,
        errors: Optional[List[Type[RPCError]]],
    ) -> Tuple[dict, dict]:
        errors_fields: Dict[str, dict] = {}
        errors_models: Dict[str, ModelField] = {}
        if not errors:
            return errors_fields, errors_models

        errors_fields = {
            error.__fields__["id"].default: {
                "description": error.__doc__ or error.__fields__["reason"].default,
            }
            for error in errors
            if error.__fields__["id"].default
        }
        errors_models = {
            error.__fields__["id"].default: ModelField(
                name=error.__name__,
                type_=error,
                class_validators=None,
                model_config=BaseConfig,
            )
            for error in errors
            if error.__fields__["id"].default
        }

        return errors_fields, errors_models
